from ...extra.eg cimport Example
from ...structs cimport ConstantsC
from ...typedefs cimport len_t
from ..forward cimport dot_plus
from ..forward cimport ELU

import numpy as np


def call_dot_plus__ELU(float[:] top_acts, float[:] btm_acts, float[:] W, float[:] bias):
    eg = Example([btm_acts.shape[0], top_acts.shape[0]])
    eg.set_input(btm_acts)
    #dot_plus__ELU(eg.c.fwd_state, eg.c.bwd_state[0], &W[0], <const len_t*>eg.c.widths,
    #              0, 3, &hp)
    dot_plus(&top_acts[0], &bias[0], bias.shape[0], &btm_acts[0], btm_acts.shape[0], &W[0])
    ELU(&top_acts[0], bias.shape[0])


#def call_d_ELU__(float[:] top_acts, float[:] btm_acts, float[:] W, float[:] bias):
#    cdef const ConstantsC hp
#    d_ELU__dot(gradient,
#        bwd, averages, W, fwd, shape, 3, 3, &hp)



from __future__ import print_function

from ._funcs_shim import *
import numpy as np
np.random.seed(0)
from numpy.testing import assert_allclose

from hypothesis import given, assume
from hypothesis.strategies import tuples, lists, integers, floats
from hypothesis.extra.numpy import arrays


def lengths(lo=1, hi=10):
    return integers(min_value=lo, max_value=hi)


def shapes(min_rows=1, max_rows=100, min_cols=1, max_cols=100):
    return tuples(lengths(lo=min_rows, hi=max_rows), lengths(lo=min_cols, hi=max_cols))


def ndarrays_of_shape(shape, lo=-1000.0, hi=1000.0):
    return arrays('float32', shape=shape, elements=floats(min_value=lo, max_value=hi))
    

def ndarrays(min_len=0, max_len=10, min_val=-10000000.0, max_val=1000000.0):
    return lengths(lo=1, hi=2).flatmap(
        lambda n: ndarrays_of_shape(n, lo=min_val, hi=max_val))


def matrices(min_rows=1, max_rows=10, min_cols=1, max_cols=10,
             min_value=-10000000.0, max_value=1000000.0):
    return shapes(min_rows=min_rows, max_rows=max_rows,
                  min_cols=min_cols, max_cols=max_cols).flatmap(
                        lambda mn: ndarrays_of_shape(mn, lo=min_value, hi=max_value))


def positive_ndarrays(min_len=0, max_len=10, max_val=100000.0):
    return ndarrays(min_len=min_len, max_len=max_len, min_val=0, max_val=max_val)


def negative_ndarrays(min_len=0, max_len=10, min_val=-100000.0):
    return ndarrays(min_len=min_len, max_len=max_len, min_val=min_val, max_val=-1e-10)

#return lengths.flatmap(
    #    lambda n: arrays('float32', shape=(n,), elements=elements))

def parse_layer(layer_data):
    # Get the first row, excluding the first column
    x = layer_data[0,1:]
    # Get the first column, excluding the first row
    # .ascontiguousarray is support important here!!!!
    b = np.ascontiguousarray(layer_data[1:,0], dtype='float32')
    # Slice out the row and the column used for the X and the bias
    W = layer_data[1:,1:]
    assert x.ndim == 1
    assert b.ndim == 1
    assert b.shape[0] == W.shape[0]
    assert x.shape[0] == W.shape[1]
    assume(not np.isnan(W.sum()))
    assume(not np.isnan(x.sum()))
    assume(not np.isnan(b.sum()))
    assume(not any(np.isinf(val) for val in W.flatten()))
    assume(not any(np.isinf(val) for val in x))
    assume(not any(np.isinf(val) for val in b))
    return x, b, W
 

def split_row(layer_data):
    return (layer_data[0,:], layer_data[:,:])


@given(matrices(min_rows=2, min_cols=2, max_rows=4, max_cols=4,
                min_value=-1.0, max_value=1.0))
def test_dot_plus(layer_data):
    x, b, W = parse_layer(layer_data)
    my_result = np.zeros(shape=(len(b),), dtype='float32')
    call_dot_plus(my_result,
        x, W.flatten(), b, len(b), len(x))
    numpy_result = W.dot(x) + b
    assert_allclose(numpy_result, my_result, atol=1e-5)


@given(matrices(min_rows=2, min_cols=2, max_rows=4, max_cols=4,
                min_value=-1.0, max_value=1.0))
def test_d_dot(layer_data):
    x, top_diff, W = parse_layer(layer_data)
    my_result = np.zeros(shape=(len(x),), dtype='float32')
    call_d_dot(my_result,
        top_diff, W.flatten(), W.shape[0], W.shape[1])
    numpy_result = W.T.dot(top_diff)
    assert_allclose(numpy_result, my_result, rtol=1e-05)


@given(matrices(min_rows=2, min_cols=2, max_rows=4, max_cols=4,
                min_value=-1.0, max_value=1.0))
def test_elu(layer_data):
    x, bias, W = parse_layer(layer_data)
 
    numpy_x = x.copy()
    call_ELU(x, len(x))

    gold = _chainer_elu(numpy_x)
    assert_allclose(x, gold, rtol=1e-05)

def _chainer_elu(x):
    y = x.copy()
    neg_indices = x < 0
    y[neg_indices] = 1.0 * (np.exp(y[neg_indices]) - 1)
    return y


@given(matrices(min_rows=2, min_cols=2, max_rows=4, max_cols=4,
                min_value=-1.0, max_value=1.0))
def test_d_elu(layer_data):
    def _get_gold(gx, x):
        # Code taken from Chainer, but modified to match our API
        neg_indices = x < 0
        gx[neg_indices] *= x[neg_indices] + 1.0
        return gx

    delta, layer_data = split_row(layer_data)
    signal_out, layer_data = split_row(layer_data)
    
    gold = _get_gold(delta.copy(), signal_out.copy())
   
    call_d_ELU(delta, signal_out, len(signal_out))

    assert_allclose(delta, gold, rtol=1e-05)


#def test_normalize():
#    pass



# cython: profile=True
# cython: cdivision=True
# cython: infer_types=True
cimport cython
from libc.string cimport memcpy, memset

from ..typedefs cimport len_t
from ..typedefs cimport idx_t

from ..linalg cimport MatMat, MatVec, VecVec, Vec, sqrtf


DEF EPS = 0.00000001 
DEF ALPHA = 1.0


@cython.cdivision(True)
cdef void vanilla_sgd(float* weights, float* moments, float* gradient,
        len_t nr_weight,const ConstantsC* hp) nogil:
    '''
    Update weights with vanilla SGD
    '''
    # Add the derivative of the L2-loss to the gradient
    if hp.r != 0:
        VecVec.add_i(gradient,
            weights, hp.r, nr_weight)
    VecVec.add_i(weights,
        gradient, -hp.e, nr_weight)
    memset(gradient,
        0, sizeof(gradient[0]) * nr_weight)


@cython.cdivision(True)
cdef void sgd_cm(float* weights, float* momentum, float* gradient,
        len_t nr_weight,const ConstantsC* hp) nogil:
    '''
    Update weights with SGD and classical momentum
    '''
    # Add the derivative of the L2-loss to the gradient
    if hp.r != 0:
        VecVec.add_i(gradient,
            weights, hp.r, nr_weight)
    Vec.mul_i(momentum, hp.m, nr_weight)
    VecVec.add_i(momentum,
        gradient, hp.e, nr_weight)
    VecVec.add_i(weights,
        momentum, -1.0, nr_weight)
    memset(gradient,
        0, sizeof(gradient[0]) * nr_weight)


@cython.cdivision(True)
cdef void adam(
    float* weights, float* moments, float* gradient,
        len_t nr_weight, const ConstantsC* hp) nogil:
    cdef float beta1 = 0.90
    cdef float beta2 = 0.999
    # Add the derivative of the L2-loss to the gradient
    cdef idx_t i
    if hp.r != 0:
        VecVec.add_i(gradient,
            weights, hp.r, nr_weight)
    mom1 = moments
    Vec.mul_i(mom1,
        beta1, nr_weight)
    VecVec.add_i(mom1,
        gradient, 1-beta1, nr_weight)
    mom2 = &moments[nr_weight]
    for i in range(nr_weight):
        mom2[i] = (beta2 * mom2[i]) + ((1-beta2) * gradient[i] * gradient[i])
    # More efficient version, from the paper
    cdef float a_t = hp.e * sqrtf(1-beta2**hp.t) / (1-beta1**hp.t)
    for i in range(nr_weight):
        weights[i] -= a_t * (mom1[i] / (sqrtf(mom2[i]) + EPS))
    memset(gradient, 0, sizeof(gradient[0]) * nr_weight)


@cython.cdivision(True)
cdef void adagrad(
    float* weights, float* moments, float* gradient,
        len_t nr_weight, const ConstantsC* hp) nogil:
    # Add the derivative of the L2-loss to the gradient
    cdef int i
    if hp.r != 0:
        VecVec.add_i(gradient,
            weights, hp.r, nr_weight)
    VecVec.add_pow_i(moments,
        gradient, 2.0, nr_weight)
    for i in range(nr_weight):
        gradient[i] *= hp.e / (sqrtf(moments[i]) + EPS)
    # Make the (already scaled) update
    VecVec.add_i(weights,
        gradient, -1.0, nr_weight)
    memset(gradient, 0, sizeof(gradient[0]) * nr_weight)


@cython.cdivision(True)
cdef void adadelta(float* weights, float* momentum, float* gradient,
        len_t nr_weight, const ConstantsC* hp) nogil:
    cdef float alpha = 0.90
    # Add the derivative of the L2-loss to the gradient
    cdef int i
    if hp.r != 0:
        VecVec.add_i(gradient,
            weights, hp.r, nr_weight)
    avg = momentum
    Vec.mul_i(avg,
        alpha, nr_weight)
    for i in range(nr_weight):
        avg[i] += (1-alpha) * gradient[i] * gradient[i]
    step = &momentum[nr_weight]
    for i in range(nr_weight):
        gradient[i] *= sqrtf(step[i] + EPS) / sqrtf(avg[i] + EPS)
    VecVec.add_i(weights,
        gradient, -1.0, nr_weight)
    Vec.mul_i(step,
        alpha, nr_weight)
    memset(gradient, 0, sizeof(gradient[0]) * nr_weight)



cdef class Example:
    def __init__(self, int nr_class=0, int nr_atom=0,
            int nr_feat=0, widths=None, Pool mem=None):
        if mem is None:
            mem = Pool()
        self.mem = mem
        self.fill_features(0, nr_feat)
        self.fill_atoms(0, nr_atom)
        self.fill_is_valid(1, nr_class)
        self.fill_scores(0, nr_class)
        self.fill_costs(0, nr_class)
        if widths is not None:
            self.c.nr_layer = len(widths)
            self.c.widths = <int*>self.mem.alloc(
                sizeof(self.c.widths[0]), len(widths))
            self.c.fwd_state = <weight_t**>self.mem.alloc(
                sizeof(self.c.fwd_state[0]), len(widths))
            self.c.bwd_state = <weight_t**>self.mem.alloc(
                sizeof(self.c.bwd_state[0]), len(widths))
            for i, width in enumerate(widths):
                self.c.widths[i] = width
                self.c.fwd_state[i] = <weight_t*>self.mem.alloc(
                    sizeof(self.c.fwd_state[i][0]), width)
                self.c.bwd_state[i] = <weight_t*>self.mem.alloc(
                    sizeof(self.c.bwd_state[i][0]), width)

    cpdef int fill_features(self, int value, int nr_feat) except -1:
        if self.c.features == NULL:
            self.c.features = <FeatureC*>self.mem.alloc(
                sizeof(self.c.features[0]), nr_feat)
            self.c.nr_feat = nr_feat
        elif self.c.nr_feat < nr_feat:
            self.c.features = <FeatureC*>self.mem.realloc(self.c.features,
                sizeof(self.c.features[0]) * nr_feat)
        for i in range(nr_feat):
            self.c.features[i].i = value
            self.c.features[i].key = value
            self.c.features[i].value = value
        self.c.nr_feat = nr_feat

    cpdef int fill_atoms(self, atom_t value, int nr_atom) except -1:
        if self.c.atoms == NULL:
            self.c.atoms = <atom_t*>self.mem.alloc(sizeof(self.c.atoms[0]), nr_atom)
            self.c.nr_atom = nr_atom
        elif self.c.nr_atom < nr_atom:
            self.c.atoms = <atom_t*>self.mem.realloc(self.c.atoms,
                sizeof(self.c.atoms[0]) * nr_atom)
        for i in range(nr_atom):
            self.c.atoms[i] = value
        self.c.nr_atom = nr_atom

    cpdef int fill_scores(self, weight_t value, int nr_class) except -1:
        if self.c.scores == NULL:
            self.c.scores = <weight_t*>self.mem.alloc(sizeof(self.c.scores[0]), nr_class)
            self.c.nr_class = nr_class
        elif self.c.nr_class < nr_class:
            self.c.scores = <weight_t*>self.mem.realloc(self.c.scores,
                sizeof(self.c.scores[0]) * nr_class)
        for i in range(nr_class):
            self.c.scores[i] = value
        self.c.nr_class = nr_class

    cpdef int fill_is_valid(self, int value, int nr_class) except -1:
        if self.c.is_valid == NULL:
            self.c.is_valid = <int*>self.mem.alloc(sizeof(self.c.is_valid[0]), nr_class)
            self.c.nr_class = nr_class
        elif self.c.nr_class < nr_class:
            self.c.is_valid = <int*>self.mem.realloc(self.c.is_valid,
                sizeof(self.c.is_valid[0]) * nr_class)
        for i in range(nr_class):
            self.c.is_valid[i] = value
        self.c.nr_class = nr_class
   
    cpdef int fill_costs(self, weight_t value, int nr_class) except -1:
        if self.c.costs == NULL:
            self.c.costs = <weight_t*>self.mem.alloc(sizeof(self.c.costs[0]), nr_class)
            self.c.nr_class = nr_class
        elif self.c.nr_class < nr_class:
            self.c.costs = <weight_t*>self.mem.realloc(self.c.costs,
                sizeof(self.c.costs[0]) * nr_class)
        for i in range(nr_class):
            self.c.costs[i] = value
        self.c.nr_class = nr_class
    
    def reset(self):
        self.fill_features(0, self.c.nr_feat)
        self.fill_atoms(0, self.c.nr_atom)
        self.fill_scores(0, self.c.nr_class)
        self.fill_costs(0, self.c.nr_class)
        self.fill_is_valid(1, self.c.nr_class)
        for i in range(self.c.nr_layer):
            memset(self.c.fwd_state[i],
                0, sizeof(self.c.fwd_state[i][0]) * self.c.widths[i])
            memset(self.c.bwd_state[i],
                0, sizeof(self.c.bwd_state[i][0]) * self.c.widths[i])
   
    def set_input(self, input_):
        if len(input_) > self.c.widths[0]:
            lengths = (len(input_), self.c.widths[0])
            raise IndexError("Cannot set %d elements to input of length %d" % lengths)
        cdef int i
        cdef float value
        for i, value in enumerate(input_):
            self.c.fwd_state[0][i] = value

    property features:
        def __get__(self):
            for i in range(self.nr_feat):
                yield self.c.features[i]
        def __set__(self, features):
            cdef weight_t value
            cdef int slot
            if isinstance(features, dict):
                feats_dict = features
                features = []
                for key, value in feats_dict.items():
                    if isinstance(key, int):
                        slot = 0
                    else:
                        slot, key = key
                    features.append((slot, key, value))
            self.nr_feat = len(features)
            cdef feat_t feat
            for i, (slot, feat, value) in enumerate(features):
                self.c.features[i] = FeatureC(i=slot, key=feat, value=value)

    property scores:
        def __get__(self):
            return [self.c.scores[i] for i in range(self.c.nr_class)]
        def __set__(self, scores):
            is_valid = list(scores)
            if len(scores) < self.nr_class:
                self.fill_scores(0, self.nr_class)
            else:
                self.nr_class = len(scores)
            for i, score in enumerate(scores):
                self.c.scores[i] = score

    property is_valid:
        def __get__(self):
            return [self.c.is_valid[i] for i in range(self.c.nr_class)]
        def __set__(self, is_valid):
            is_valid = list(is_valid)
            if len(is_valid) < self.nr_class:
                self.fill_is_valid(1, self.nr_class)
            else:
                self.nr_class = len(is_valid)
            for i, is_valid in enumerate(is_valid):
                self.c.is_valid[i] = is_valid

    property costs:
        def __get__(self):
            return [self.c.costs[i] for i in range(self.c.nr_class)]
        def __set__(self, costs):
            costs = list(costs)
            if len(costs) < self.nr_class:
                self.fill_costs(0, self.nr_class)
            else:
                self.nr_class = len(costs)
            cdef int i
            cdef weight_t cost
            for i, cost in enumerate(costs):
                self.c.costs[i] = cost

    property guess:
        def __get__(self):
            return VecVec.arg_max_if_true(self.c.scores, self.c.is_valid, self.c.nr_class)

    property best:
        def __get__(self):
            return VecVec.arg_max_if_zero(self.c.scores, self.c.costs, self.c.nr_class)
    
    property cost:
        def __get__(self):
            return self.c.costs[self.guess]
        def __set__(self, weight_t value):
            self.c.costs[self.guess] = value
   
    property nr_class:
        def __get__(self):
            return self.c.nr_class
        def __set__(self, int nr_class):
            if self.c.nr_class != nr_class:
                if self.c.scores is not NULL:
                    self.mem.free(self.c.scores)
                if self.c.is_valid is not NULL:
                    self.mem.free(self.c.is_valid)
                if self.c.costs is not NULL:
                    self.mem.free(self.c.costs)
                self.c.scores = <weight_t*>self.mem.alloc(
                    sizeof(self.c.scores[0]), nr_class)
                self.c.is_valid = <int*>self.mem.alloc(
                    sizeof(self.c.is_valid[0]), nr_class)
                self.c.costs = <weight_t*>self.mem.alloc(
                    sizeof(self.c.costs[0]), nr_class)
                self.c.nr_class = nr_class
 
    property nr_atom:
        def __get__(self):
            return self.c.nr_atom
        def __set__(self, int nr_atom):
            if self.c.nr_atom != nr_atom:
                if self.c.atoms is not NULL:
                    self.mem.free(self.c.atoms)
            self.c.atoms = <atom_t*>self.mem.alloc(
                sizeof(self.c.atoms[0]), nr_atom)
            self.c.nr_atom = nr_atom
 
    property nr_feat:
        def __get__(self):
            return self.c.nr_feat
        def __set__(self, int nr_feat):
            if self.c.nr_feat != nr_feat:
                if self.c.features is not NULL:
                    self.mem.free(self.c.features)
                self.c.features = <FeatureC*>self.mem.alloc(
                    sizeof(self.c.features[0]), nr_feat)
                self.c.nr_feat = nr_feat

    property loss:
        def __get__(self):
            return 1 - self.c.scores[self.best]

    def activation(self, int i, int j):
        # TODO: Find a way to do this better!
        return self.c.fwd_state[i][j]

    def delta(self, int i, int j):
        # TODO: Find a way to do this better!
        return self.c.bwd_state[i][j]



from libc.stdint cimport int16_t, int, int32_t, uint64_t
from preshed.maps cimport MapStruct

from .typedefs cimport len_t, idx_t, atom_t


include "compile_time_constants.pxi"


ctypedef void (*do_update_t)(
    float* weights,
    float* momentum,
    float* gradient,
        len_t nr,
        const ConstantsC* hp
) nogil


ctypedef void (*do_feed_fwd_t)(
    float** fwd,
    float* averages,
        const float* W,
        const len_t* shape,
        int nr_below,
        int nr_above,
        const ConstantsC* hp
) nogil
 

ctypedef void (*do_feed_bwd_t)(
    float* G,
    float** bwd,
    float* averages,
        const float* W,
        const float* const* fwd,
        const len_t* shape,
        int nr_above,
        int nr_below,
        const ConstantsC* hp
) nogil


# Alias this, so that it matches our naming scheme
ctypedef MapStruct MapC


cdef struct ConstantsC:
    float a
    float b
    float c
    float d
    float e
    float g
    float h
    float i
    float j
    float k
    float l
    float m
    float n
    float o
    float p
    float q
    float r
    float s
    float t
    float u
    float w
    float x
    float y
    float z


cdef struct EmbedC:
    MapC** weights
    MapC** momentum
    idx_t* offsets
    len_t* lengths
    len_t nr


cdef struct NeuralNetC:
    do_feed_fwd_t feed_fwd
    do_feed_bwd_t feed_bwd
    do_update_t update

    len_t* widths
    float* weights
    float* gradient
    float* momentum

    float** averages
    
    EmbedC embed

    len_t nr_layer
    len_t nr_weight
    len_t nr_node

    ConstantsC hp


cdef struct ExampleC:
    int* is_valid
    float* costs
    uint64_t* atoms
    FeatureC* features
    float* scores

    float** fwd_state
    float** bwd_state
    int* widths

    int nr_class
    int nr_atom
    int nr_feat
    int nr_layer


cdef packed struct SparseArrayC:
    int32_t key
    float val


cdef struct FeatureC:
    int i
    uint64_t key
    float value


cdef struct SparseAverageC:
    SparseArrayC* curr
    SparseArrayC* avgs
    SparseArrayC* times


cdef struct TemplateC:
    int[MAX_TEMPLATE_LEN] indices
    int length
    atom_t[MAX_TEMPLATE_LEN] atoms



# cython: profile=True
# cython: cdivision=True

cimport cython
from libc.stdint cimport int32_t
from libc.string cimport memcpy
from cymem.cymem cimport Pool

ctypedef float weight_t

include "compile_time_constants.pxi"


# Copied from Shane Legg's Tokyo
#cdef extern from "cblas.h":
#    enum CBLAS_ORDER:     CblasRowMajor, CblasColMajor
#    enum CBLAS_TRANSPOSE: CblasNoTrans, CblasTrans, CblasConjTrans
#    enum CBLAS_UPLO:      CblasUpper, CblasLower
#    enum CBLAS_DIAG:      CblasNonUnit, CblasUnit
#    enum CBLAS_SIDE:      CblasLeft, CblasRight
#
#    # BLAS level 1 routines
#
#    void cblas_sswap(int M, float  *x, int incX, float  *y, int incY) nogil
#    void cblas_sscal(int N, float  alpha, float  *x, int incX) nogil
#    void cblas_scopy(int N, float  *x, int incX, float  *y, int incY) nogil
#    void cblas_saxpy(int N, float  alpha, float  *x, int incX, float  *y, int incY ) nogil
#    float cblas_sdot(int N, float  *x, int incX, float  *y, int incY ) nogil
#    float cblas_snrm2(int N, float  *x, int incX) nogil
#    float cblas_sasum(int N, float  *x, int incX) nogil
#    int cblas_isamax(int N, float  *x, int incX) nogil
#
#    # BLAS level 2 routines
#    void cblas_sgemv(CBLAS_ORDER Order, CBLAS_TRANSPOSE TransA, int M, int N,
#                                 float  alpha, float  *A, int lda, float  *x, int incX,
#                                 float  beta, float  *y, int incY) nogil
#
#    void cblas_sger(CBLAS_ORDER Order, int M, int N, float  alpha, float  *x,
#                                int incX, float  *y, int incY, float  *A, int lda) nogil
#
#    # BLAS level 3 routines
#    void cblas_sgemm(CBLAS_ORDER Order, CBLAS_TRANSPOSE TransA,
#                                 CBLAS_TRANSPOSE TransB, int M, int N, int K,
#                                 float  alpha, float  *A, int lda, float  *B, int ldb,
#                                 float  beta, float  *C, int ldc) nogil
#


cdef extern from "math.h" nogil:
    float expf(float x)
    float sqrtf(float x)


cdef class Matrix:
    cdef readonly Pool mem
    cdef weight_t* data
    cdef readonly int32_t nr_row
    cdef readonly int32_t nr_col


cdef class Vec:
    @staticmethod    
    cdef inline int arg_max(const weight_t* scores, const int n_classes) nogil:
        cdef int i
        cdef int best = 0
        cdef weight_t mode = scores[0]
        for i in range(1, n_classes):
            if scores[i] > mode:
                mode = scores[i]
                best = i
        return best

    @staticmethod
    cdef inline weight_t max(const weight_t* x, int32_t nr) nogil:
        if nr == 0:
            return 0
        cdef int i
        cdef weight_t mode = x[0]
        for i in range(1, nr):
            if x[i] > mode:
                mode = x[i]
        return mode

    @staticmethod
    cdef inline weight_t sum(const weight_t* vec, int32_t nr) nogil:
        cdef int i
        cdef weight_t total = 0
        for i in range(nr):
            total += vec[i]
        return total

    @staticmethod
    cdef inline void add(weight_t* output, const weight_t* x,
            weight_t inc, int32_t nr) nogil:
        memcpy(output, x, sizeof(output[0]) * nr)
        Vec.add_i(output, inc, nr)

    @staticmethod
    cdef inline void add_i(weight_t* vec, weight_t inc, int32_t nr) nogil:
        cdef int i
        for i in range(nr):
            vec[i] += inc

    @staticmethod
    cdef inline void mul(weight_t* output, const weight_t* vec, weight_t scal,
            int32_t nr) nogil:
        memcpy(output, vec, sizeof(output[0]) * nr)
        Vec.mul_i(output, scal, nr)

    @staticmethod
    cdef inline void mul_i(weight_t* vec, const weight_t scal, int32_t nr) nogil:
        cdef int i
        if USE_BLAS:
            cblas_sscal(nr, scal, vec, 1)
        else:
            for i in range(nr):
                vec[i] *= scal

    @staticmethod
    cdef inline void pow(weight_t* output, const weight_t* vec, weight_t scal,
            int32_t nr) nogil:
        memcpy(output, vec, sizeof(output[0]) * nr)
        Vec.pow_i(output, scal, nr)

    @staticmethod
    cdef inline void pow_i(weight_t* vec, const weight_t scal, int32_t nr) nogil:
        cdef int i
        for i in range(nr):
            vec[i] **= scal

    @staticmethod
    @cython.cdivision(True)
    cdef inline void div(weight_t* output, const weight_t* vec, weight_t scal,
            int32_t nr) nogil:
        memcpy(output, vec, sizeof(output[0]) * nr)
        Vec.div_i(output, scal, nr)

    @staticmethod
    @cython.cdivision(True)
    cdef inline void div_i(weight_t* vec, const weight_t scal, int32_t nr) nogil:
        cdef int i
        for i in range(nr):
            vec[i] /= scal

    @staticmethod
    cdef inline void exp(weight_t* output, const weight_t* vec, int32_t nr) nogil:
        memcpy(output, vec, sizeof(output[0]) * nr)
        Vec.exp_i(output, nr)

    @staticmethod
    cdef inline void exp_i(weight_t* vec, int32_t nr) nogil:
        cdef int i
        for i in range(nr):
            vec[i] = expf(vec[i])

    @staticmethod
    cdef inline void reciprocal_i(weight_t* vec, int32_t nr) nogil:
        cdef int i
        for i in range(nr):
            vec[i] = 1.0 / vec[i]


cdef class VecVec:
    @staticmethod
    cdef inline void add(weight_t* output,
                         const weight_t* x, 
                         const weight_t* y,
                         weight_t scale,
                         int32_t nr) nogil:
        memcpy(output, x, sizeof(output[0]) * nr)
        VecVec.add_i(output, y, scale, nr)
   
    @staticmethod
    cdef inline void add_i(float* x, 
                           const float* y,
                           float scale,
                           int32_t nr) nogil:
        cdef int i
        if USE_BLAS:
            cblas_saxpy(nr, scale, y, 1, x, 1)
        else:
            for i in range(nr):
                x[i] += y[i] * scale
 
    @staticmethod
    cdef inline void add_pow(weight_t* output,
            const weight_t* x, const weight_t* y, weight_t power, int32_t nr) nogil:
        memcpy(output, x, sizeof(output[0]) * nr)
        VecVec.add_pow_i(output, y, power, nr)

   
    @staticmethod
    cdef inline void add_pow_i(weight_t* x, 
            const weight_t* y, weight_t power, int32_t nr) nogil:
        cdef int i
        for i in range(nr):
            x[i] += y[i] ** power
 
    @staticmethod
    cdef inline void mul(weight_t* output,
            const weight_t* x, const weight_t* y, int32_t nr) nogil:
        memcpy(output, x, sizeof(output[0]) * nr)
        VecVec.mul_i(output, y, nr)
   
    @staticmethod
    cdef inline void mul_i(weight_t* x, 
            const weight_t* y, int32_t nr) nogil:
        cdef int i
        for i in range(nr):
            x[i] *= y[i]

 
    @staticmethod
    cdef inline weight_t dot(
            const weight_t* x, const weight_t* y, int32_t nr) nogil:
        cdef int i
        cdef weight_t total = 0
        for i in range(nr):
            total += x[i] * y[i]
        return total
 
    @staticmethod
    cdef inline int arg_max_if_true(
            const weight_t* scores, const int* is_valid, const int n_classes) nogil:
        cdef int i
        cdef int best = 0
        cdef weight_t mode = -900000
        for i in range(n_classes):
            if is_valid[i] and scores[i] > mode:
                mode = scores[i]
                best = i
        return best

    @staticmethod
    cdef inline int arg_max_if_zero(
            const weight_t* scores, const weight_t* costs, const int n_classes) nogil:
        cdef int i
        cdef int best = 0
        cdef weight_t mode = -900000
        for i in range(n_classes):
            if costs[i] == 0 and scores[i] > mode:
                mode = scores[i]
                best = i
        return best


cdef class MatVec:
    @staticmethod
    cdef inline void mul(weight_t* output,
                         const weight_t* mat,
                         const weight_t* vec,
                         int32_t nr_row, int32_t nr_col) nogil:
        memcpy(output, mat, sizeof(output[0]) * nr_row * nr_col)
        MatVec.mul_i(output, vec, nr_row, nr_col)

    @staticmethod
    cdef inline void mul_i(weight_t* mat,
                           const weight_t* vec,
                           int32_t nr_row, int32_t nr_col) nogil:
        cdef int i, row, col
        for i in range(nr_row):
            row = i * nr_col
            for col in range(nr_col):
                mat[row + col] *= vec[col]

    @staticmethod
    cdef inline void dot(weight_t* output,
                         const weight_t* mat,
                         const weight_t* vec,
                         int32_t nr_row, int32_t nr_col) nogil:
        cdef int i, row, col
        if USE_BLAS:
            cblas_sgemv(
                CblasRowMajor,
                CblasNoTrans,
                nr_row,
                nr_col,
                1.0,
                mat,
                nr_col,
                vec,
                1,
                0.0,
                output,
                1
            )
        else:
            for i in range(nr_row):
                output[i] = 0
                row = i * nr_col
                for col in range(nr_col):
                    output[i] += mat[row + col] * vec[col]

    @staticmethod
    cdef inline void T_dot(weight_t* output,
                             const weight_t* mat,
                             const weight_t* vec,
                             int32_t nr_row,
                             int32_t nr_col) nogil:
        cdef int i, row, col
        if USE_BLAS:
            cblas_sgemv(
                CblasRowMajor,
                CblasTrans,
                nr_row,
                nr_col,
                1.0,
                mat,
                nr_col,
                vec,
                1,
                0.0,
                output,
                1
            )
        else:
            for row in range(nr_row):
                for col in range(nr_col):
                    output[col] += vec[row] * mat[(row * nr_col) + col]


cdef class MatMat:
    @staticmethod
    cdef inline void add(weight_t* output,
                         const weight_t* x,
                         const weight_t* y,
                         int32_t nr_row, int32_t nr_col) nogil:
        memcpy(output, x, sizeof(output[0]) * nr_row * nr_col)
        MatMat.add_i(output, y, nr_row, nr_col)

    @staticmethod
    cdef inline void add_i(weight_t* x,
                           const weight_t* y,
                           int32_t nr_row, int32_t nr_col) nogil:
        cdef int i, row, col
        for i in range(nr_row):
            row = i * nr_col
            for col in range(nr_col):
                x[row + col] += y[row + col]

    @staticmethod
    cdef inline void mul(weight_t* output,
                         const weight_t* x,
                         const weight_t* y,
                         int32_t nr_row, int32_t nr_col) nogil:
        memcpy(output, x, sizeof(output[0]) * nr_row * nr_col)
        MatMat.mul_i(output, y, nr_row, nr_col)

    @staticmethod
    cdef inline void mul_i(weight_t* x,
                           const weight_t* y,
                           int32_t nr_row, int32_t nr_col) nogil:
        cdef int i, row, col
        for i in range(nr_row):
            row = i * nr_col
            for col in range(nr_col):
                x[row + col] *= y[row + col]

    @staticmethod 
    cdef inline void add_outer_i(weight_t* mat,
                                 const weight_t* x,
                                 const weight_t* y,
                                 int32_t nr_row,
                                 int32_t nr_col) nogil:
        cdef int i, j, row
        for i in range(nr_row):
            row = i * nr_col
            for j in range(nr_col):
                mat[row + j] += x[i] * y[j]



# cython: profile=True
# cython: cdivision=True
# cython: infer_types=True
cimport cython

from ..typedefs cimport len_t
from ..typedefs cimport idx_t

from ..linalg cimport MatMat, MatVec, VecVec, Vec, sqrtf, expf


DEF EPS = 0.00000001 
DEF ALPHA = 1.0


cdef void dot_plus__ELU(float** fwd, float* averages,
        const float* W, const len_t* shape, int nr_below, int nr_above,
        const ConstantsC* hp) nogil:
    bias = W + shape[1] * shape[0]
    dot_plus(fwd[1],
        bias, shape[1], fwd[0], shape[0], W)
    # Apply non-linearity
    if nr_above >= 2:
        ELU(fwd[1],
            shape[1])
    else:
        softmax(fwd[1],
            shape[1])
 

cdef void dot_plus__ReLu(float** fwd, float* averages,
        const float* W, const len_t* shape, int nr_below, int nr_above,
        const ConstantsC* hp) nogil:
    bias = W + shape[1] * shape[0]
    dot_plus(fwd[1],
        bias, shape[1], fwd[0], shape[0], W)
    # Apply non-linearity
    if nr_above >= 2:
        ReLu(fwd[1],
            shape[1])
    else:
        softmax(fwd[1],
            shape[1])
 

cdef void dot_plus__residual__ELU(float** fwd, float* averages,
        const float* W, const len_t* shape, int nr_below, int nr_above,
        const ConstantsC* hp) nogil:
    bias = W + shape[1] * shape[0]
    dot_plus(fwd[1],
        bias, shape[1], fwd[0], shape[0], W)
    if nr_below >= 1 and shape[-1] == shape[1]:
        VecVec.add_i(fwd[1],
            fwd[-1], 1.0, shape[1])
    # Apply non-linearity
    if nr_above >= 2:
        ELU(fwd[1],
            shape[1])
    else:
        softmax(fwd[1],
            shape[1])


cdef void dot__normalize__dot_plus__ELU(float** fwd, float* averages,
        const float* W, const len_t* shape, int nr_before, int nr_above,
        const ConstantsC* hp) nogil:
    # Read the bias and gamma terms from the weights data
    bias = W + shape[1] * shape[0]
    # Gamma is the normalization rescaling weights
    gamma = bias + shape[1]
    # Read the E(x) and Var(x) estimates from 'averages'
    Ex = averages
    Vx = &averages[shape[1]]
    # We write our output in fwd[1][0...n]
    # An imporant intermediary result is the batch normed activation, which
    # we compute in fwd[1][n...2n], and preserve for the backward pass.
    x_norm = fwd[1] + shape[1]

    MatVec.dot(fwd[1],
        W, fwd[0], shape[1], shape[0])
    normalize(x_norm, Ex, Vx,
        fwd[1], shape[1], hp.a, hp.t)
    VecVec.mul(fwd[1],
        x_norm, gamma, shape[1])
    VecVec.add_i(fwd[1],
        bias, 1.0, shape[1])
    # Apply non-linearity
    if nr_above >= 2:
        ELU(fwd[1],
            shape[1])
    else:
        softmax(fwd[1],
            shape[1])


cdef void dot_plus(float* out,
        const float* bias, len_t nr_out,
        const float* x, len_t nr_in,
        const float* W) nogil:
    MatVec.dot(out,
        W, x, nr_out, nr_in)
    cdef float one = 1.0
    if bias is not NULL:
        VecVec.add_i(out,
            bias, one, nr_out)


cdef void ELU(float* out, len_t nr_out) nogil:
    cdef idx_t i
    for i in range(nr_out):
        if out[i] < 0:
            out[i] = ALPHA * (expf(out[i]) - 1)


cdef void ReLu(float* out, len_t nr_out) nogil:
    cdef idx_t i
    for i in range(nr_out):
        if out[i] < 0:
            out[i] = 0


cdef void softmax(float* out, len_t nr_out) nogil:
    #w = exp(w - max(w))
    Vec.add_i(out,
        -Vec.max(out, nr_out), nr_out)
    Vec.exp_i(out,
        nr_out)
    #w = w / sum(w)
    cdef float norm = Vec.sum(out, nr_out)
    if norm != 0:
        Vec.div_i(out,
            norm, nr_out)


cdef void normalize(float* x_norm, float* Ex, float* Vx,
        const float* x, len_t nr_x, float alpha, float time) nogil:
    # Upd EMA estimate of mean and variance
    # See eq at the end of this:
    # http://nfs-uxsup.csx.cam.ac.uk/~fanf2/hermes/doc/antiforgery/stats.pdf
    cdef idx_t i
    cdef float diff
    cdef float incr
    cdef float one = 1.0
    for i in range(nr_x):
        diff = x[i] - Ex[i]
        incr = alpha * diff
        Vx[i] = (one - alpha) * (Vx[i] + diff * incr)
        Ex[i] += incr
    # Normalize
    if time < 100:
        for i in range(nr_x):
            x_norm[i] = x[i]
    else:
        for i in range(nr_x):
            if (x[i] - Ex[i]) == 0:
                x_norm[i] = 0
            else:
                x_norm[i] = (x[i] - Ex[i]) / sqrtf(Vx[i] + EPS)



# cython: profile=True
# cython: cdivision=True
# cython: infer_types=True
from __future__ import print_function

from libc.string cimport memmove, memset, memcpy

cimport cython

from cymem.cymem cimport Pool
from preshed.maps cimport map_init as Map_init
from preshed.maps cimport map_set as Map_set
from preshed.maps cimport map_get as Map_get
from preshed.maps cimport map_iter as Map_iter
from preshed.maps cimport key_t

from ..typedefs cimport weight_t, atom_t, feat_t
from ..typedefs cimport len_t, idx_t
from ..linalg cimport MatMat, MatVec, VecVec, Vec
from ..structs cimport MapC
from ..structs cimport NeuralNetC
from ..structs cimport ExampleC
from ..structs cimport FeatureC
from ..structs cimport EmbedC
from ..structs cimport ConstantsC
from ..structs cimport do_update_t

from ..extra.eg cimport Example

from .solve cimport vanilla_sgd, sgd_cm, adam, adagrad

from .solve cimport adam
from .solve cimport adadelta
from .solve cimport adagrad
from .solve cimport vanilla_sgd
from .forward cimport dot_plus__ELU
from .forward cimport dot_plus__ReLu
from .forward cimport dot_plus__residual__ELU
from .forward cimport dot__normalize__dot_plus__ELU
from .backward cimport d_ELU__dot__normalize__dot
from .backward cimport d_ELU__dot
from .backward cimport d_ReLu__dot

from .backward cimport d_log_loss

from libc.string cimport memcpy
from libc.math cimport isnan, sqrt

import random
import numpy


DEF USE_BATCH_NORM = False


cdef class NN:
    @staticmethod
    cdef int nr_weight(int nr_out, int nr_in) nogil:
        return nr_out * nr_in + nr_out

    @staticmethod
    cdef void init(
        NeuralNetC* nn,
        Pool mem,
            widths,
            embed=None,
            update_step='adam',
            float eta=0.005,
            float eps=1e-6,
            float mu=0.9,
            float rho=1e-4,
            float alpha=0.5
    ) except *:
        if update_step == 'sgd':
            nn.update = vanilla_sgd
        elif update_step == 'sgd_cm':
            nn.update = sgd_cm
        elif update_step == 'adadelta':
            nn.update = adadelta
        elif update_step == 'adagrad':
            nn.update = adagrad
        else:
            nn.update = adam
        nn.feed_fwd = dot_plus__ELU
        nn.feed_bwd = d_ELU__dot

        nn.hp.t = 0
        nn.hp.a = alpha
        nn.hp.r = rho
        nn.hp.m = mu
        nn.hp.e = eta

        nn.nr_layer = len(widths)
        nn.widths = <len_t*>mem.alloc(nn.nr_layer, sizeof(widths[0]))
        nn.averages = <float**>mem.alloc(nn.nr_layer, sizeof(void*))
        cdef int i
        for i, width in enumerate(widths):
            nn.widths[i] = width
            nn.averages[i] = <float*>mem.alloc(width*4, sizeof(nn.averages[i][0]))
        nn.nr_weight = 0
        nn.nr_node = 0
        for i in range(nn.nr_layer-1):
            nn.nr_weight += NN.nr_weight(nn.widths[i+1], nn.widths[i])
            nn.nr_node += nn.widths[i]
        nn.weights = <float*>mem.alloc(nn.nr_weight, sizeof(nn.weights[0]))
        nn.gradient = <float*>mem.alloc(nn.nr_weight, sizeof(nn.weights[0]))
        nn.momentum = <float*>mem.alloc(nn.nr_weight * 2, sizeof(nn.weights[0]))
        
        if embed is not None:
            vector_widths, features = embed
            Embedding.init(&nn.embed, mem, vector_widths, features)

        W = nn.weights
        for i in range(nn.nr_layer-2):
            he_uniform_initializer(W,
                nn.widths[i+1] * nn.widths[i])
            if USE_BATCH_NORM:
                he_uniform_initializer(W+nn.widths[i+1] * nn.widths[i] + nn.widths[i+1],
                    nn.widths[i+1] * nn.widths[i])
            W += NN.nr_weight(nn.widths[i+1], nn.widths[i])
        if USE_BATCH_NORM:
            i = nn.nr_layer-2
            he_uniform_initializer(W+nn.widths[i+1] * nn.widths[i] + nn.widths[i+1],
                nn.widths[i+1] * nn.widths[i])
    
    @staticmethod
    cdef void train_example(NeuralNetC* nn, Pool mem, ExampleC* eg) except *:
        nn.hp.t += 1
        Embedding.insert_missing(mem, &nn.embed,
            eg.features, eg.nr_feat)
        Embedding.set_input(eg.fwd_state[0],
            eg.features, eg.nr_feat, &nn.embed)
        NN.forward(eg.scores, eg.fwd_state,
            nn)
        NN.backward(eg.bwd_state, nn.gradient,
            eg.fwd_state, eg.costs, nn)
        nn.update(nn.weights, nn.momentum, nn.gradient,
            nn.nr_weight, &nn.hp)
        if eg.nr_feat != 0:
            Embedding.fine_tune(&nn.embed, nn.gradient,
                eg.bwd_state[0], nn.widths[0], eg.features, eg.nr_feat,
                &nn.hp, nn.update)
    
    @staticmethod
    cdef void forward(float* scores, float** fwd, const NeuralNetC* nn) nogil:
        cdef const float* W = nn.weights
        for i in range(nn.nr_layer-1):
            nn.feed_fwd(&fwd[i], nn.averages[i+1],
                W, &nn.widths[i], i, nn.nr_layer-(i+1), &nn.hp)
            W += NN.nr_weight(nn.widths[i+1], nn.widths[i])
        memcpy(scores,
            fwd[nn.nr_layer-1], sizeof(scores[0]) * nn.widths[nn.nr_layer-1])

    @staticmethod
    cdef void backward(float** bwd, float* gradient,
            const float* const* fwd, const float* costs, const NeuralNetC* nn) nogil:
        d_log_loss(bwd[nn.nr_layer-1],
            costs, fwd[nn.nr_layer-1], nn.widths[nn.nr_layer-1])
        cdef const float* W = nn.weights + nn.nr_weight
        cdef float* G = gradient + nn.nr_weight
        for i in range(nn.nr_layer-2, -1, -1):
            W -= NN.nr_weight(nn.widths[i+1], nn.widths[i])
            G -= NN.nr_weight(nn.widths[i+1], nn.widths[i])
            nn.feed_bwd(G, &bwd[i], nn.averages[i+1],
                W, &fwd[i], &nn.widths[i], nn.nr_layer-(i+1), i, &nn.hp)


cdef class Embedding:
    cdef Pool mem
    cdef EmbedC* c

    @staticmethod
    cdef void init(EmbedC* self, Pool mem, vector_widths, features) except *: 
        assert max(features) < len(vector_widths)
        # Create tables, which may be shared between different features
        # e.g., we might have a feature for this word, and a feature for next
        # word. These occupy different parts of the input vector, but draw
        # from the same embedding table.
        self.nr = len(features)
        uniq_weights = <MapC*>mem.alloc(len(vector_widths), sizeof(MapC))
        uniq_momentum = <MapC*>mem.alloc(len(vector_widths), sizeof(MapC))
        for i, width in enumerate(vector_widths):
            Map_init(mem, &uniq_weights[i], 8)
            Map_init(mem, &uniq_momentum[i], 8)
        self.offsets = <idx_t*>mem.alloc(len(features), sizeof(len_t))
        self.lengths = <len_t*>mem.alloc(len(features), sizeof(len_t))
        self.weights = <MapC**>mem.alloc(len(features), sizeof(void*))
        self.momentum = <MapC**>mem.alloc(len(features), sizeof(void*))
        offset = 0
        for i, table_id in enumerate(features):
            self.weights[i] = &uniq_weights[table_id]
            self.momentum[i] = &uniq_momentum[table_id]
            self.lengths[i] = vector_widths[table_id]
            self.offsets[i] = offset
            offset += vector_widths[table_id]

    @staticmethod
    cdef void set_input(float* out,
            const FeatureC* features, len_t nr_feat, const EmbedC* embed) nogil:
        for feat in features[:nr_feat]:
            emb = <const float*>Map_get(embed.weights[feat.i], feat.key)
            if emb is not NULL:
                VecVec.add_i(&out[embed.offsets[feat.i]], 
                    emb, feat.value, embed.lengths[feat.i])

    @staticmethod
    cdef void insert_missing(Pool mem, EmbedC* embed,
            const FeatureC* features, len_t nr_feat) except *:
        for feat in features[:nr_feat]:
            if feat.i >= embed.nr:
                continue
            emb = <float*>Map_get(embed.weights[feat.i], feat.key)
            if emb is NULL:
                emb = <float*>mem.alloc(embed.lengths[feat.i], sizeof(emb[0]))
                he_uniform_initializer(emb, embed.lengths[feat.i])
                Map_set(mem, embed.weights[feat.i],
                    feat.key, emb)
                # Need 2x length for momentum. Need to centralize this somewhere =/
                mom = <float*>mem.alloc(embed.lengths[feat.i] * 2, sizeof(mom[0]))
                Map_set(mem, embed.momentum[feat.i],
                    feat.key, mom)
    
    @staticmethod
    cdef inline void fine_tune(EmbedC* layer, weight_t* fine_tune,
            const weight_t* delta, int nr_delta, const FeatureC* features, int nr_feat,
            const ConstantsC* hp, do_update_t do_update) nogil:
        for feat in features[:nr_feat]:
            # Reset fine_tune, because we need to modify the gradient
            memcpy(fine_tune, delta, sizeof(float) * nr_delta)
            weights = <weight_t*>Map_get(layer.weights[feat.i], feat.key)
            gradient = &fine_tune[layer.offsets[feat.i]]
            mom = <float*>Map_get(layer.momentum[feat.i], feat.key)
            # None of these should ever be null
            do_update(weights, mom, gradient,
                layer.lengths[feat.i], hp)


cdef class NeuralNet:
    def __init__(self, widths, embed=None,
                 weight_t eta=0.005, weight_t eps=1e-6, weight_t mu=0.2,
                 weight_t rho=1e-4, weight_t alpha=0.99,
                 update_step='adam'):
        self.mem = Pool()
        NN.init(&self.c, self.mem, widths, embed, update_step,
                eta, eps, mu, rho, alpha)
        self.eg = Example(nr_class=self.nr_class, widths=self.widths)

    def predict_example(self, Example eg):
        if eg.c.nr_feat != 0:
            Embedding.insert_missing(self.mem, &self.c.embed,
                eg.c.features, eg.c.nr_feat)
            Embedding.set_input(eg.c.fwd_state[0],
                eg.c.features, eg.c.nr_feat, &self.c.embed)
        NN.forward(eg.c.scores, eg.c.fwd_state,
            &self.c)
        return eg

    def predict_dense(self, features):
        cdef Example eg = Example(nr_class=self.nr_class, widths=self.widths)
        cdef weight_t value
        for i, value in enumerate(features):
            eg.c.fwd_state[0][i] = value
        return self.predict_example(eg)

    def predict_sparse(self, features):
        cdef Example eg = self.Example(features)
        return self.predict_example(eg)
    
    def train_dense(self, features, y):
        cdef Example eg = Example(nr_class=self.nr_class, widths=self.widths)
        cdef weight_t value 
        for i, value in enumerate(features):
            eg.c.fwd_state[0][i] = value
        eg.costs = y
        NN.train_example(&self.c, self.mem, &eg.c)
        return eg
  
    def train_sparse(self, features, label):
        cdef Example eg = self.Example(features, label=label)
        NN.train_example(&self.c, self.mem, &eg.c)
        return eg
   
    def train_example(self, Example eg):
        NN.train_example(&self.c, self.mem, &eg.c)
        return eg
 
    def Example(self, input_, label=None):
        if isinstance(input_, Example):
            return input_
        cdef Example eg = Example(nr_class=self.nr_class, widths=self.widths)
        eg.features = input_
        if label is not None:
            if isinstance(label, int):
                eg.costs = [i != label for i in range(eg.nr_class)]
            else:
                eg.costs = label
        return eg

    property use_batch_norm:
        def __get__(self):
            return USE_BATCH_NORM

    property weights:
        def __get__(self):
            return [self.c.weights[i] for i in range(self.c.nr_weight)]
        def __set__(self, weights):
            assert len(weights) == self.c.nr_weight
            for i, weight in enumerate(weights):
                self.c.weights[i] = weight

    property averages:
        def __get__(self):
            for i, width in enumerate(self.widths):
                yield [self.c.averages[i][j] for j in range(width*4)]

    property layers:
        def __get__(self):
            weights = list(self.weights)
            start = 0
            for i in range(self.c.nr_layer-1):
                nr_w = self.widths[i] * self.widths[i+1]
                nr_bias = self.widths[i] * self.widths[i+1] + self.widths[i+1]
                W = weights[start:start+nr_w]
                bias = weights[start+nr_w:start+nr_w+nr_bias]
                yield W, bias
                start = start + NN.nr_weight(self.widths[i+1], self.widths[i])

    property widths:
        def __get__(self):
            return tuple(self.c.widths[i] for i in range(self.c.nr_layer))

    property layer_l1s:
        def __get__(self):
            for W, bias in self.layers:
                w_l1 = sum(abs(w) for w in W) / len(W)
                bias_l1 = sum(abs(w) for w in W) / len(bias)
                yield w_l1, bias_l1

    property gradient:
        def __get__(self):
            return [self.c.gradient[i] for i in range(self.c.nr_weight)]

    property l1_gradient:
        def __get__(self):
            cdef int i
            cdef weight_t total = 0.0
            for i in range(self.c.nr_weight):
                if self.c.gradient[i] < 0:
                    total -= self.c.gradient[i]
                else:
                    total += self.c.gradient[i]
            return total / self.c.nr_weight

    property embeddings:
        def __get__(self):
            cdef int i = 0
            cdef int j = 0
            cdef int k = 0
            cdef key_t key
            cdef void* value
            for i in range(self.c.embed.nr):
                j = 0
                while Map_iter(self.c.embed.weights[i], &j, &key, &value):
                    emb = <weight_t*>value
                    yield key, [emb[k] for k in range(self.c.embed.lengths[i])]

    property nr_layer:
        def __get__(self):
            return self.c.nr_layer
    property nr_weight:
        def __get__(self):
            return self.c.nr_weight
    property nr_class:
        def __get__(self):
            return self.c.widths[self.c.nr_layer-1]
    property nr_in:
        def __get__(self):
            return self.c.widths[0]

    property eta:
        def __get__(self):
            return self.c.hp.e
        def __set__(self, eta):
            self.c.hp.e = eta
    property rho:
        def __get__(self):
            return self.c.hp.rho
        def __set__(self, rho):
            self.c.hp.r = rho
    property eps:
        def __get__(self):
            return self.c.hp.p
        def __set__(self, eps):
            self.c.hp.p = eps


cdef void he_normal_initializer(float* weights, int fan_in, int n) except *:
    # See equation 10 here:
    # http://arxiv.org/pdf/1502.01852v1.pdf
    values = numpy.random.normal(loc=0.0, scale=numpy.sqrt(2.0 / float(fan_in)), size=n)
    cdef float value
    for i, value in enumerate(values):
        weights[i] = value


cdef void he_uniform_initializer(float* weights, int n) except *:
    # See equation 10 here:
    # http://arxiv.org/pdf/1502.01852v1.pdf
    values = numpy.random.randn(n) * numpy.sqrt(2.0/n)
    cdef float value
    for i, value in enumerate(values):
        weights[i] = value


cdef void constant_initializer(float* weights, float value, int n) nogil:
    for i in range(n):
        weights[i] = value



from ..structs cimport FeatureC
from ..structs cimport ConstantsC

from ..typedefs cimport len_t
from ..typedefs cimport idx_t


cdef void dot_plus__ELU(float** fwd, float* averages,
        const float* W, const len_t* shape, int nr_below, int nr_above,
        const ConstantsC* hp) nogil
 

cdef void dot_plus__ReLu(float** fwd, float* averages,
        const float* W, const len_t* shape, int nr_below, int nr_above,
        const ConstantsC* hp) nogil
 

cdef void dot_plus__residual__ELU(float** fwd, float* averages,
        const float* W, const len_t* shape, int nr_below, int nr_above,
        const ConstantsC* hp) nogil


cdef void dot__normalize__dot_plus__ELU(float** fwd, float* averages,
        const float* W, const len_t* shape, int nr_before, int nr_above,
        const ConstantsC* hp) nogil


cdef void dot_plus(float* out,
        const float* bias, len_t nr_out,
        const float* x, len_t nr_in,
        const float* W) nogil
  

cdef void softmax(float* out, len_t nr_out) nogil


cdef void ELU(float* out, len_t nr_out) nogil



from __future__ import division

import pytest
import pickle
import io
import tempfile

from thinc.linear.avgtron import AveragedPerceptron
from thinc.extra.eg import Example


def assert_near_eq(float1, float2):
    assert abs(float1 - float2) < 0.001


def test_basic():
    nr_class = 3
    model = AveragedPerceptron(((1,), (2,), (3,), (4,), (5,)))
    instances = [
        (1, {1: 1, 3: -5}),
        (2, {2: 4, 3: 5})
    ]
    for clas, feats in instances:
        eg = Example(nr_class)
        eg.features = feats
        model(eg)
        eg.costs = [i != clas for i in range(nr_class)]
        model.update(eg)
    eg = Example(nr_class)
    eg.features = {1: 2, 2: 1}
    model(eg)
    assert eg.guess == 2
    eg = Example(nr_class)
    eg.features = {0: 2, 2: 1}
    model(eg)
    assert eg.scores[1] == 0
    eg = Example(nr_class)
    eg.features = {1: 2, 2: 1}
    model(eg)
    assert eg.scores[2] > 0
    eg = Example(nr_class)
    eg.features = {1: 2, 1: 1}
    model(eg)
    assert eg.scores[1] > 0
    eg = Example(nr_class)
    eg.features = {0: 3, 3: 1}
    model(eg)
    assert eg.scores[1] < 0 
    eg = Example(nr_class)
    eg.features = {0: 3, 3: 1}
    model(eg)
    assert eg.scores[2] > 0 


@pytest.fixture
def instances():
    instances = [
        [
            (1, {1: -1, 2: 1}),
            (2, {1: 5, 2: -5}),
            (3, {1: 3, 2: -3}),
        ],
        [
            (1, {1: -1, 2: 1}),
            (2, {1: -1, 2: 2}),
            (3, {1: 3, 2: -3})
        ],
        [
            (1, {1: -1, 2: 2}),
            (2, {1: 5, 2: -5}), 
            (3, {4: 1, 5: -7, 2: 1})
        ]
    ]
    return instances

@pytest.fixture
def model(instances):
    templates = []
    for batch in instances:
        for _, feats in batch:
            for key in feats:
                templates.append((key,))
    templates = tuple(set(templates))
    model = AveragedPerceptron(templates)
    for batch in instances:
        model.time += 1
        for clas, feats in batch:
            for key, value in feats.items():
                model.update_weight(key, clas, value)
    return model

def get_score(nr_class, model, feats, clas):
    eg = Example(nr_class)
    eg.features = feats
    eg.costs = [i != clas for i in range(nr_class)]
    model(eg)
    return eg.scores[clas]


def get_scores(nr_class, model, feats):
    eg = Example(nr_class)
    eg.features = feats
    model(eg)
    return list(eg.scores)


def test_averaging(model):
    model.end_training()
    nr_class = 4
    # Feature 1
    assert_near_eq(get_score(nr_class, model, {1: 1}, 1), sum([-1, -2, -3]) / 3.0)
    assert_near_eq(get_score(nr_class, model, {1: 1}, 2), sum([5, 4, 9]) / 3.0)
    assert_near_eq(get_score(nr_class, model, {1: 1}, 3), sum([3, 6, 6]) / 3.0)
    # Feature 2
    assert_near_eq(get_score(nr_class, model, {2: 1}, 1), sum([1, 2, 4]) / 3.0)
    assert_near_eq(get_score(nr_class, model, {2: 1}, 2), sum([-5, -3, -8]) / 3.0)
    assert_near_eq(get_score(nr_class, model, {2: 1}, 3), sum([-3, -6, -5]) / 3.0)
    # Feature 3 (absent)
    assert_near_eq(get_score(nr_class, model, {3: 1}, 1), 0)
    assert_near_eq(get_score(nr_class, model, {3: 1}, 2), 0)
    assert_near_eq(get_score(nr_class, model, {3: 1}, 3), 0)
    # Feature 4
    assert_near_eq(get_score(nr_class, model, {4: 1}, 1), sum([0, 0, 0]) / 3.0)
    assert_near_eq(get_score(nr_class, model, {4: 1}, 2), sum([0, 0, 0]) / 3.0)
    assert_near_eq(get_score(nr_class, model, {4: 1}, 3), sum([0, 0, 1]) / 3.0)
    # Feature 5
    assert_near_eq(get_score(nr_class, model, {5: 1}, 1), sum([0, 0, 0]) / 3.0)
    assert_near_eq(get_score(nr_class, model, {5: 1}, 2), sum([0, 0, 0]) / 3.0)
    assert_near_eq(get_score(nr_class, model, {5: 1}, 3), sum([0, 0, -7]) / 3.0)


def test_dump_load(model):
    loc = tempfile.mkstemp()[1]
    model.end_training()
    model.dump(loc)
    string = open(loc, 'rb').read()
    assert string
    new_model = AveragedPerceptron([(1,), (2,), (3,), (4,)])
    nr_class = 5
    assert get_scores(nr_class, model, {1: 1, 3: 1, 4: 1}) != \
           get_scores(nr_class, new_model, {1:1, 3:1, 4:1})
    assert get_scores(nr_class, model, {2:1, 5:1}) != \
            get_scores(nr_class, new_model, {2:1, 5:1})
    assert get_scores(nr_class, model, {2:1, 3:1, 4:1}) != \
           get_scores(nr_class, new_model, {2:1, 3:1, 4:1})
    new_model.load(loc)
    assert get_scores(nr_class, model, {1:1, 3:1, 4:1}) == \
           get_scores(nr_class, new_model, {1:1, 3:1, 4:1})
    assert get_scores(nr_class, model, {2:1, 5:1}) == \
           get_scores(nr_class, new_model, {2:1, 5:1})
    assert get_scores(nr_class, model, {2:1, 3:1, 4:1}) == \
           get_scores(nr_class, new_model, {2:1, 3:1, 4:1})


## TODO: Need a test that exercises multiple lines. Example bug:
## in gather_weights, don't increment f_i per row, only per feature
## (so overwrite some lines we're gathering)



from thinc.structs cimport NeuralNetC
from thinc.extra.eg cimport Example
from cymem.cymem cimport Pool


cdef class NeuralNet:
    cdef readonly Pool mem
    cdef readonly Example eg
    cdef NeuralNetC c





from ..structs cimport ConstantsC

from ..typedefs cimport len_t
from ..typedefs cimport idx_t


cdef void vanilla_sgd(float* weights, float* moments, float* gradient,
        len_t nr_weight,const ConstantsC* hp) nogil


cdef void sgd_cm(float* weights, float* moments, float* gradient,
        len_t nr_weight, const ConstantsC* hp) nogil


cdef void adam(
    float* weights, float* moments, float* gradient,
        len_t nr_weight, const ConstantsC* hp) nogil

 
cdef void adagrad(
    float* weights, float* moments, float* gradient,
        len_t nr_weight, const ConstantsC* hp) nogil
 

cdef void adadelta(float* weights, float* momentum, float* gradient,
        len_t nr_weight, const ConstantsC* hp) nogil



from __future__ import division, print_function

import pytest
import pickle
import io
import numpy as np
import random
from numpy.testing import assert_allclose

from thinc.neural.nn import NeuralNet
from thinc.extra.eg import Example

np.random.seed(2)
random.seed(0)


def test_create():
    model = NeuralNet((4, 8, 3))

    assert model.nr_in == 4
    assert model.nr_class == 3
    assert model.nr_layer == 3
    assert model.widths == (4, 8, 3)
    assert model.mem is not None

    with pytest.raises(AttributeError) as excinfo:
        model.mem = None

def test_fwd_bias():
    model = NeuralNet((2, 2), rho=0.0)
    
    model.weights = [1.0] * model.nr_weight

    eg = model.predict_dense([0, 0])
    assert eg.nr_class == 2
    assert_allclose(eg.scores, [0.5, 0.5])

    # Set bias for class 0
    syn = [1.,1.,1.,1.]
    bias = [100000.,1.]
    gamma = [0,0]
    if model.nr_weight == len(syn) + len(bias):
        model.weights = syn + bias
        assert model.weights == list(syn + bias)
    else:
        model.weights = syn + bias + gamma
        assert model.weights == list(syn + bias + gamma)
    eg = model.predict_dense([0, 0])
    assert_allclose(eg.scores, [1.0, 0.0])

    # Set bias for class 1
    if model.nr_weight == 6:
        model.weights = syn + [1.,10000.0]
    else:
        model.weights = syn + [1.,10000.0] + gamma
    eg = model.predict_dense([0,0])
    assert_allclose(eg.scores, [0.0, 1.0])

    # Set bias for both
    if model.nr_weight == 6:
        model.weights = syn + [10000.0,10000.0]
    else:
        model.weights = syn + [10000.0,10000.0] + gamma
    eg = model.predict_dense([0,0])
    assert_allclose(eg.scores, [0.5, 0.5])


def test_fwd_linear():
    model = NeuralNet((2,2), rho=0.0, alpha=0.5)
    assert model.nr_class == 2
    assert model.widths == (2, 2)

    syn = [1.,0.,0.,1.]
    bias = [0.,0.]
    gamma = [1.,1.]
    if model.use_batch_norm:
        model.weights = syn+bias+gamma
    else:
        model.weights = syn+bias
    
    ff = [0,0]
    tf = [1,0]
    ft = [0,1]
    tt = [1,1]
    eg = model.predict_dense(ff)

    assert_allclose(eg.scores, [0.5, 0.5])

    eg = model.predict_dense(ft)
    
    assert_allclose(eg.scores, [ 0.26894142,  0.73105858])
    assert_allclose([sum(eg.scores)], [1.0])

    eg = model.predict_dense(tf)
    assert_allclose(eg.scores, [0.73105858, 0.26894142])
    assert_allclose(sum(eg.scores), [1.0])

    eg = model.predict_dense(tt)

    assert_allclose(eg.scores, [0.5, 0.5])


def test_xor_manual():
    model = NeuralNet((2,2,2), rho=0.0)
    assert model.nr_class == 2
    assert model.widths == (2, 2, 2)

    # Make a network that detects X-or
    # It should output 0 if inputs are 0,0 or 1,1 and 1 if inputs are 0,1 or 1,0
    # A linear model can't do this!
    # 
    # What we do is create two intermediate predictors, for 0,1 and 1,0
    # These predictors rely on a bias towards 0. The non-linearity is essential
    # Then our output layer can detect either of these signals firing
    #
    # 0,0 --> neither fire
    # 0,1 --> A0 fires
    # 1,0 --> A1 fires
    # 1,1 --> neither fire
    #

    if not model.use_batch_norm:
        model.weights = np.asarray([
                    [4.0, -10.0],   # A.0*in.0, A.0*in.1
                    [-10.0, 5.0], # A.1*in.0, A.1*in.1
                    [0.0, 0.0],     # A.0 bias, A.1 bias
                    [-10.0, -10.0],  # out.0*A.0, out.0*A.1
                    [10.0, 10.0],   # out.1*A.0, out.1*A.1
                    [10.0, -10.0],   # out.0 bias, out.1 bias
                ]).flatten()
    else:
        model.weights = np.asarray([
                    [4.0, -10.0],   # A.0*in.0, A.0*in.1
                    [-10.0, 5.0], # A.1*in.0, A.1*in.1
                    [0.0, 0.0],     # A.0 bias, A.1 bias
                    [1.0, 1.0],     # A.0 gamma, A.1 gamma
                    [-10.0, -10.0],  # out.0*A.0, out.0*A.1
                    [10.0, 10.0],   # out.1*A.0, out.1*A.1
                    [10.0, -10.0],   # out.0 bias, out.1 bias
                    [1.0, 1.0],     # out.0 gamma, out.1 gamma
                ]).flatten()

    ff = [0,0]
    tf = [1,0]
    ft = [0,1]
    tt = [1,1]

    eg = model.predict_dense(ff)
    assert eg.scores[0] > 0.99
 
    eg = model.predict_dense(tt)
    assert eg.scores[0] > 0.99
    
    eg = model.predict_dense(tf)
    assert eg.scores[1] > 0.99

    eg = model.predict_dense(ft)
    assert eg.scores[1] > 0.99
 

@pytest.fixture
def xor_data():
    ff = np.asarray([0.,0.], dtype='f')
    tf = np.asarray([1.,0.], dtype='f')
    ft = np.asarray([0.,1.], dtype='f')
    tt = np.asarray([1.,1.], dtype='f')
    costs0 = np.asarray([0., 1.], dtype='f')
    costs1 = np.asarray([1., 0.], dtype='f')
    return [(ff, 0, costs0), (tf, 1, costs1), (ft, 1, costs1), (tt, 0, costs0)]


@pytest.fixture
def or_data():
    ff = np.asarray([0.,0.], dtype='f')
    tf = np.asarray([1.,0.], dtype='f')
    ft = np.asarray([0.,1.], dtype='f')
    tt = np.asarray([1.,1.], dtype='f')
    costs0 = np.asarray([0., 1.], dtype='f')
    costs1 = np.asarray([1., 0.], dtype='f')
    return [(ff, 0, costs0), (tf, 1, costs1), (ft, 1, costs1), (tt, 1, costs1)]


@pytest.fixture
def bias_data():
    ff = np.asarray([0.,0.], dtype='f')
    costs0 = np.asarray([0., 1.], dtype='f')
    costs1 = np.asarray([1., 0.], dtype='f')
    def gen_random():
        if random.random() > 0.7:
            yield (ff, 0, costs0)
        else:
            yield (ff, 1, costs1)
    return gen_random


def test_linear_bias(bias_data):
    '''Test that a linear model can learn a bias.'''
    model = NeuralNet((2, 2), rho=0.0, eta=0.1, eps=1e-4, update_step='sgd')

    assert model.nr_in == 2
    assert model.nr_class == 2
    assert model.nr_layer == 2
    
    if not model.use_batch_norm:
        bias0, bias1 = model.weights[-2:]
    else:
        bias0, bias1 = model.weights[-4:-2]

    assert bias0 == 0
    assert bias1 == 0
    for _ in range(100):
        for feats, label, costs in bias_data():
            eg = model.train_dense(feats, costs)
    if not model.use_batch_norm:
        bias0, bias1 = model.weights[-2:]
    else:
        bias0, bias1 = model.weights[-4:-2]
    assert bias1 > bias0
    acc = 0.0
    total = 0
    for i in range(20):
        for features, label, costs in bias_data():
            eg = model.predict_dense(features)
            assert costs[label] == 0
            acc += eg.scores[label] > 0.5
            total += 1
    assert (acc/total) > 0.5
    assert (acc/total) < 1.0


def test_deep_bias(bias_data):
    '''Test that a deep model can learn a bias.'''
    model = NeuralNet((2,2,2,2,2,2,2, 2), rho=0.0, eta=0.1, eps=1e-4, update_step='adadelta')

    assert model.nr_in == 2
    assert model.nr_class == 2
    assert model.nr_layer > 2
    
    if not model.use_batch_norm:
        bias0, bias1 = model.weights[-2:]
    else:
        bias0, bias1 = model.weights[-4:-2]
    assert bias0 == 0
    assert bias1 == 0
    for _ in range(20):
        for feats, label, costs in bias_data():
            eg = model.train_dense(feats, costs)
    if not model.use_batch_norm:
        bias0, bias1 = model.weights[-2:]
    else:
        bias0, bias1 = model.weights[-4:-2]
 
    assert bias1 > bias0
    acc = 0.0
    total = 0
    for i in range(20):
        for features, label, costs in bias_data():
            eg = model.predict_dense(features)
            assert costs[label] == 0
            acc += eg.scores[label] > 0.5
            total += 1
    assert (acc/total) > 0.5
    assert (acc/total) < 1.0


def test_learn_linear(or_data):
    '''Test that a linear model can learn OR.'''
    # Need high eta on this sort of toy problem, or learning takes forever!
    model = NeuralNet((2, 2), rho=0.0, eta=0.1, eps=1e-4, update_step='sgd',
                       alpha=0.8)

    assert model.nr_in == 2
    assert model.nr_class == 2
    assert model.nr_layer == 2
    
    # It takes about this many iterations, with the settings above.
    for _ in range(900):
        for feats, label, costs in or_data:
            eg = model.train_dense(feats, costs)
        random.shuffle(or_data)
    for avg in model.averages:
        print(avg)
    acc = 0.0
    for features, label, costs in or_data:
        eg = model.predict_dense(features)
        assert costs[label] == 0
        acc += eg.scores[label] > 0.5
    assert acc == len(or_data)


def test_mlp_learn_linear(or_data):
    '''Test that with a hidden layer, we can still learn OR'''
    # Need high eta on this sort of toy problem, or learning takes forever!
    model = NeuralNet((2, 3, 2), rho=0.0, eta=0.5, eps=1e-4,
                      update_step='sgd')

    assert model.nr_in == 2
    assert model.nr_class == 2
    assert model.nr_layer == 3
    
    # Keep this set low, so that we see that the hidden layer allows the function
    # to be learned faster than the linear model
    for _ in range(50):
        for feats, label, costs in or_data:
            batch = model.train_dense(feats, costs)
        random.shuffle(or_data)
    acc = 0.0
    for features, label, costs in or_data:
        eg = model.predict_dense(features)
        assert costs[label] == 0
        acc += eg.scores[label] > 0.5
    assert acc == len(or_data)


def test_xor_gradient(xor_data):
    '''Test that after each update, we move towards the correct label.'''
    model = NeuralNet((2, 2, 2), rho=0.0, eta=0.1, update_step='adadelta')
    assert model.nr_in == 2
    assert model.nr_class == 2
    assert model.nr_layer == 3
    
    for _ in range(500):
        for i, (features, label, costs) in enumerate(xor_data):
            prev = list(model.predict_dense(features).scores)
            assert_allclose([sum(prev)], [1.0])
            tmp = model.train_dense(features, costs)
            eg = model.predict_dense(features)
            assert (prev[label] <= eg.scores[label] or \
                    prev[label] == eg.scores[label] == 1.0)


@pytest.mark.xfail
def test_xor_eta(xor_data):
    '''Test that a higher learning rate causes loss to decrease faster.'''
    small_eta_model = NeuralNet((2, 10,2), rho=0.0, eta=0.000001, update_step='sgd')
    normal_eta_model = NeuralNet((2, 10,2), rho=0.0, eta=0.1, update_step='sgd')
    small_eta_loss = 0.0
    normal_eta_loss = 0.0
    for _ in range(1000):
        for i, (features, label, costs) in enumerate(xor_data):
            small_eta_loss += small_eta_model.train_dense(features, costs).loss
            normal_eta_loss += normal_eta_model.train_dense(features, costs).loss
    assert normal_eta_loss < small_eta_loss


def test_xor_rho(xor_data):
    '''Test that higher L2 penalty causes slower learning.'''
    big_rho_model = NeuralNet((2,10,10,10,2), rho=0.8, eta=0.005, update_step='adadelta')
    normal_rho_model = NeuralNet((2, 10,10,10, 2), rho=1e-4, eta=0.005, update_step='adadelta')
    big_rho_model.weights = list(normal_rho_model.weights)
    big_rho_loss = 0.0
    normal_rho_loss = 0.0
    for _ in range(10):
        for i, (features, label, costs) in enumerate(xor_data):
            big_rho_loss += big_rho_model.train_dense(features, costs).loss
            normal_rho_loss += normal_rho_model.train_dense(features, costs).loss
    assert normal_rho_loss < (big_rho_loss * 1.1)


def test_xor_deep(xor_data):
    '''Compare 0, 1 and 3 layer networks.
    The 3 layer seems to do better, but it doesn't *have* to. But if the
    0 layer works, something's wrong!'''
    linear = NeuralNet((2,2), rho=0.0000, eta=0.1, update_step='sgd')
    small = NeuralNet((2,2,2), rho=0.0000, eta=0.1, update_step='sgd')
    big = NeuralNet((2,2,2,2,2,2), rho=0.0000, eta=0.1, update_step='sgd')
    for _ in range(1000):
        for i, (features, label, costs) in enumerate(xor_data):
            ln = linear.train_dense(features, costs)
            bg = big.train_dense(features, costs)
            sm = small.train_dense(features, costs)
        random.shuffle(xor_data)

    linear_loss = 0.0
    small_loss = 0.0
    big_loss = 0.0
    for i, (features, label, costs) in enumerate(xor_data):
        linear_loss += 1 - linear.predict_dense(features).scores[label]
        small_loss += 1 - small.predict_dense(features).scores[label]
        big_loss += 1 - big.predict_dense(features).scores[label]
    assert big_loss < 0.5
    assert small_loss < 2.0
    assert linear_loss > 1.9
 

def test_model_widths_sgd(or_data):
    '''Test different model widths'''
    narrow = NeuralNet((2,4,2), rho=0.0, eta=0.01, update_step='sgd')
    wide = NeuralNet((2,20,2), rho=0.0, eta=0.01, update_step='sgd')
    assert wide.nr_weight > narrow.nr_weight
    narrow_loss = 0.0
    wide_loss = 0.0
    for _ in range(20):
        for i, (features, label, costs) in enumerate(or_data):
            narrow_loss += narrow.train_dense(features, costs).loss
            wide_loss += wide.train_dense(features, costs).loss
        random.shuffle(or_data)
    # We don't know that the extra width is better, but it shouldn't be
    # *much* worse
    assert wide_loss < (narrow_loss * 1.1)
    # It also shouldn't be the same!
    assert wide_loss != narrow_loss


def test_embedding():
    model = NeuralNet((10,4,2), embed=((5,), (0,0)), rho=0.0, eta=0.005)
    assert model.nr_in == 10
    eg = model.Example({(0, 1): 2.5})
    model.predict_example(eg)
    assert eg.activation(0, 0) != 0
    assert eg.activation(0, 1) != 0
    assert eg.activation(0, 2) != 0
    assert eg.activation(0, 3) != 0
    assert eg.activation(0, 4) != 0
    
    eg = model.Example({(1, 1867): 0.5})
    model.predict_example(eg)
    assert eg.activation(0, 0) == 0.0
    assert eg.activation(0, 1) == 0.0
    assert eg.activation(0, 2) == 0.0
    assert eg.activation(0, 3) == 0.0
    assert eg.activation(0, 4) == 0.0
    assert eg.activation(0, 5) != 0.0
    assert eg.activation(0, 6) != 0.0
    assert eg.activation(0, 7) != 0.0
    assert eg.activation(0, 8) != 0.0
    assert eg.activation(0, 9) != 0.0
    

def test_sparse_backprop_single():
    model = NeuralNet((2, 2, 2), embed=((2,), (0,)), update_step='sgd', eta=0.1)
    x = {(0, 1): 4.0}
    y = (0, 1)
    b1 = model.train_sparse(x, y)
    b2 = model.train_sparse(x, y)
    b3 = model.train_sparse(x, y)
    b4 = model.train_sparse(x, y)
    b5 = model.train_sparse(x, y)
    assert b2.loss > b3.loss or b2.loss == 0.0
    assert b3.loss > b4.loss or b3.loss == 0.0
    assert b4.loss > b5.loss or b4.loss == 0.0


def f2s(fs):
    return ' '.join('%.3f' % val for val in fs)


def test_sparse_backprop():
    def train_batch(model, xy):
        x, y = xy
        loss = 0.0
        for x,y in zip(X, Y):
            eg = model.train_sparse(x, y)
            loss += eg.loss
        return loss

    model = NeuralNet((10, 2, 2), embed=((10,), (0,)), rho=0.0, eta=0.005,
                      update_step='sgd')
    X = [{(0, 1): 4.0, (0, 2): 3.0, (0, 3): 4.0, (0, 100): 1.0}, {(0, 10): 3.0,
         (0, 2): 2.0}]
    Y = [(0, 1), (1, 0)]
    b1 = train_batch(model, (X, Y))
    b2 = train_batch(model, (X, Y))
    b3 = train_batch(model, (X, Y))
    b4 = train_batch(model, (X, Y))
    b5 = train_batch(model, (X, Y))
    b6 = train_batch(model, (X, Y))
    b7 = train_batch(model, (X, Y))
    b8 = train_batch(model, (X, Y))
    b9 = train_batch(model, (X, Y))
    b10 = train_batch(model, (X, Y))

    assert b1 > b3 > b10



import theano
import theano.tensor as T



import numpy
from collections import OrderedDict, defaultdict


theano.config.floatX = 'float32'
floatX = theano.config.floatX


def th_share(w, name=''):
    return theano.shared(value=w, borrow=True, name=name)


class AvgParam(object):
    def __init__(self, numpy_data, name='?', wrapper=th_share):
        self.curr = wrapper(numpy_data, name=name+'_curr')
        self.avg = self.curr
        self.avg = wrapper(numpy_data.copy(), name=name+'_avg')
        self.step = wrapper(numpy.zeros(numpy_data.shape, numpy_data.dtype),
                            name=name+'_step')

    def updates(self, cost, timestep, momentum=0.9, eta=0.001):
        #return [(self.curr, self.curr - (eta * T.grad(cost, self.curr)))]
        step = (momentum * self.step) - T.grad(cost, self.curr)
        curr = self.curr + (eta * step)
        alpha = (1 / timestep).clip(0.001, 0.9).astype(floatX)
        avg = ((1 - alpha) * self.avg) + (alpha * curr)
        return [(self.curr, curr), (self.step, step), (self.avg, avg)]


def feed_layer(activation, weights, bias, input_):
    return activation(T.dot(input_, weights) + bias)


def L2(L2_reg, *weights):
    return L2_reg * sum((w ** 2).sum() for w in weights)


def L1(L1_reg, *weights):
    return L1_reg * sum(abs(w).sum() for w in weights)


def relu(x):
    return x * (x > 0)


def _init_weights(n_in, n_out):
    rng = numpy.random.RandomState(1234)
    weights = numpy.asarray(
        numpy.random.normal(
            loc=0.0,
            scale=0.0001,
            size=(n_in, n_out)),
        dtype=theano.config.floatX
    )
    bias = 0.2 * numpy.ones((n_out,), dtype=theano.config.floatX)
    return [AvgParam(weights, name='W'), AvgParam(bias, name='b')]


def compile_theano_model(n_classes, n_hidden, n_in, L1_reg, L2_reg):
    costs = T.ivector('costs') 
    x = T.vector('x') 
    timestep = theano.shared(1)
    eta = T.scalar('eta').astype(floatX)

    maxent_W, maxent_b = _init_weights(n_hidden, n_classes)
    hidden_W, hidden_b = _init_weights(n_in, n_hidden)

    # Feed the inputs forward through the network
    p_y_given_x = feed_layer(
                    T.nnet.softmax,
                    maxent_W.curr,
                    maxent_b.curr,
                      feed_layer(
                        relu,
                        hidden_W.curr,
                        hidden_b.curr,
                        x))

    stabilizer = 1e-8
    cost = (
        -T.log(T.sum((p_y_given_x[0] + stabilizer) * T.eq(costs, 0)))
        + L1(L1_reg, hidden_W.curr, hidden_b.curr, x)
        + L2(L2_reg, hidden_W.curr, hidden_b.curr, x)
    )

    debug = theano.function(
        name='debug',
        inputs=[x, costs],
        outputs=[p_y_given_x, T.eq(costs, 0), p_y_given_x[0] * T.eq(costs, 0)],
    )

    train_model = theano.function(
        name='train_model',
        inputs=[x, costs, eta],
        outputs=[p_y_given_x[0], T.grad(cost, x), T.argmax(p_y_given_x, axis=1)],
        updates=(
            [(timestep, timestep + 1)] + 
             maxent_W.updates(cost, timestep, eta=eta) + 
             maxent_b.updates(cost, timestep, eta=eta) +
             hidden_W.updates(cost, timestep, eta=eta) +
             hidden_b.updates(cost, timestep, eta=eta)
        )
    )

    evaluate_model = theano.function(
        name='evaluate_model',
        inputs=[x],
        outputs=[
            feed_layer(
              T.nnet.softmax,
              maxent_W.avg,
              maxent_b.avg,
              feed_layer(
                relu,
                hidden_W.avg,
                hidden_b.avg,
                x
              )
            )[0]
        ]
    )
    return debug, train_model, evaluate_model
